#include "IniParser.h"

IniParser::IniParser()
{
}

IniParser::~IniParser()
{
}

/*
 * parses INI file
 * return true on success, false on failure
 */
bool IniParser::loadIni(const QString &fileName)
{
	list.clear();

	QFile file(fileName);
	QString section;

	if (! file.open(QIODevice::ReadOnly | QIODevice::Text))
		return false;

	while (! file.atEnd()) {

		QString line = file.readLine().simplified();

		/* we've got comment or empty line, ignore */
		if (line[0] == '#' || line.isEmpty())
			continue;

		/* we've got section line, section can't be empty */
		if (line[0] == '[') {
			line.remove('[');
			line.remove(']');
			section=line.simplified();
			if (section.isEmpty())
				goto bail;
			continue;
		}

		/* we've reached "key = val" line, need section at this point */
		if (section.isEmpty())
			goto bail;

		/* check the line and clean it */
		if (! line.contains('='))
			goto bail;
		line.remove(';');

		/* build "key" and add IniEntry */
		QString key, val;
		key = section;
		key.append(":");
		key.append(line.section('=', 0, 0).simplified());
		val = line.section('=', 1, 1).simplified();
		addEntry(key, val);
	}

	file.close();
	return true;

bail:
	file.close();
	list.clear();
	return false;
}

/*
 * dumps INI list to file
 * return true on success, false on failure
 */
bool IniParser::dumpIni(const QString &fileName)
{
	/* don't even try to open a file if "list" is empty */
	if (list.isEmpty())
		return false;

	QFile file(fileName);
	QString section, lastSection;

	if (! file.open(QIODevice::WriteOnly | QIODevice::Text))
		return false;

	QTextStream out(&file);

	for (int i = 0; i < list.size(); i++) {

		section = list.at(i).key.section(':', 0, 0);

		/* section appeared/changed, need to write section line */
		if (section != lastSection) {
			out << endl << '[' << section << ']' << endl;
			lastSection = section;
		}

		/* write line, but only if set "dumpable" */
		if (list.at(i).dumpable)
			out << list.at(i).key.section(':', 1, 1) << " = " << list.at(i).val << ';' << endl;
	}

	file.close();

	return true;
}

/*
 * adds one entry in the proper place of the list
 * returns true on success, false on improper data or duplicate
 */
bool IniParser::addEntry(const QString &key, const QString &val, bool dumpable)
{
	/* check if "key" got ':' in the middle */
	if (! key.contains(':') || key.indexOf(':') == 0 || key.indexOf(':') == key.size()-1)
		return false;

	QString keySection, section, lastSection;
	keySection = key.section(':', 0, 0);
	int i;
	bool found = false;

	for (i = 0; i < list.size(); i++) {

		/* such entry already exists */
		if (list.at(i).key == key)
			return false;

		/* we need to find proper place to put entry so they'll be sorted */
		/* check what's the "key" section */
		section = list.at(i).key.section(':', 0, 0);

		/* section changed and we already found ours section, that's the place */
		if (section != lastSection && found == true)
			break;

		/* section changed */
		if (section != lastSection)
			lastSection = section;

		/* that's our section, let's wait till it ends to put new entry at the end */
		if (section == keySection)
			found = true;

		/* we found nothing, lets just add at the end */
	}

	IniEntry e;
	e.key = key;
	e.val = val;
	e.dumpable = dumpable;

	list.insert(i, e);

	return true;
}

/*
 * modifies selected entry
 * returns true on success, false on entry not found
 */
bool IniParser::modEntry(const QString &key, const QString &val)
{
	for (int i = 0; i < list.size(); i++)
		if (list.at(i).key == key) {
			list[i].val = val;
			return true;
		}

	return false;
}

/*
 * deletes entry
 * return true on success, false on entry not found
 */
bool IniParser::delEntry(const QString &key)
{
	for (int i = 0; i < list.size(); i++)
		if (list.at(i).key == key) {
			list.removeAt(i);
			return true;
		}

	return false;
}

/*
 * returns entry value
 * returns empty string on entry not found
 */
QString IniParser::getEntry(const QString &key)
{
	for (int i = 0; i < list.size(); i++)
		if (list.at(i).key == key)
			return list.at(i).val;

	return QString();
}

/*
 * sets flags of all entries to dumpable
 */
void IniParser::setDumpable(bool dumpable)
{
	for (int i = 0; i < list.size(); i++)
		list[i].dumpable = dumpable;
}

/*
 * sets entry's dumpable flag
 * returns false on entry not found
 */
bool IniParser::setDumpable(const QString &key, bool dumpable)
{
	for (int i = 0; i < list.size(); i++)
		if (list.at(i).key == key) {
			list[i].dumpable = dumpable;
			return true;
		}

	return false;
}
