#include "FS2_Flag.h"
#include "IniParser.h"
#include "Settings.h"
#include "System.h"
#include "FilePaths.h"

createFileSingleton(Settings);

Settings::Settings()
{
}

Settings::~Settings()
{
}

bool Settings::initialize()
{
	IniParser l6Ini;
	QString val;

	/* lets create config dir and data dir inside it if they dont exist */
	if (FilePaths::getGameDir() != FilePaths::getConfigDir()) {
		if (! QDir(FilePaths::getConfigDir()).exists())
			System::mkDir(FilePaths::getConfigDir());
		if (! QDir(FilePaths::getConfigDataDir()).exists())
			System::mkDir(FilePaths::getConfigDataDir());
	}

	/* if there is no laucher6.ini we bail */
	if (! l6Ini.loadIni(FilePaths::getLauncherIni())) {
		qWarning("launcher6.ini not found, select exec using browse button.");
		return false;
	}

	val = l6Ini.getEntry("launcher:exe_filepath");
	if (! QFile::exists(val)) {
		qWarning("exec not found, select one using browse button.");
		return false;
	}

	return setExeFilePath(val);
}

bool Settings::setExeFilePath(const QString &exeFilePath)
{
	/* 
	 * the following should never happen as we select file by file
	 * dialog, the only option is to modify launcher6.ini by hand
	 */
	if (! QFile::exists(exeFilePath)) {
		qWarning("CRITICAL: exec not found.");
		return false;
	}

	/* new exeFilePath, everything that was before is superfluous */
	this -> exeFilePath = exeFilePath;
	exeFlags.clear();
	exeTypes.clear();
	easyTypes.clear();
	customFlags = "";

	/* mod info should be cleared as well */
	modName = "";
	mods.clear();
	modImage = modInfo = modWebsite = modForum = "";

	IniParser l6Ini;
	QString val;

	if (!parseFlagList())
		return false;

	/* this can fail, we don't need this file now */
	l6Ini.loadIni(FilePaths::getLauncherIni());

	QString f = l6Ini.getEntry("launcher:game_flags");
	/*
	 * either game_flags is empty or whole launcher6.ini
	 * does not exist, lets try to read cmdline_fso.cfg
	 * to fill initial flag states, we need to ignore
	 * "-mod" and one param after cause of this
	 * TODO: use eventual -mod to fill setModName in
	 * such case
	 */
	if (f.isEmpty()) {
		QFile c(FilePaths::getCmdLine());
		if(c.open(QIODevice::ReadOnly | QIODevice::Text)) {
			f = c.readLine().simplified();
			c.close();
		}
	}
	QStringList flagsList = f.split(" ");

	for (int i = 0; i < flagsList.size(); i++) {
		if (flagsList.at(i) == "-mod") {
			i++;
			continue;
		}
		setFlag(flagsList.at(i), true);
	}

	val = l6Ini.getEntry("launcher:active_mod");

	return setModName(val);
}

bool Settings::parseFlagList()
{
	int easyFlagSize;
	int flagSize;

	int numEasyFlags;
	EasyFlag *easyFlags = NULL;
	int numFlags;
	Flag *exeParams = NULL;

	QFile file(FilePaths::getFlagsLch());
	file.remove();

	QStringList params;
	params.append("-get_flags");
	/*
	 * somethings wrong here, if we change dir to ~/.fs2_open game
	 * appears to be run twice, this scares the hell out of me :-)
	 * so lets just use tmp
	 */
	QString currDir = System::chDir(FilePaths::getTempDir());
	if (currDir.isEmpty()) {
		qWarning("CRITICAL: cannot change dir to tmp.");
		return false;
	}
	if (! System::run(exeFilePath, params, true)) {
		qWarning("CRITICAL: cannot exec game to test flags.");
		return false;
	}
	file.open(QIODevice::ReadOnly);

	/*
	 * for unix we dont need to wait for game to fill flags.lch,
	 * we waitpid() for it so the run() quits when the game quits,
	 * dont know how about M$ though so leaving this for now
	 */
	int wait = 1000;
	while (!file.isOpen() && wait) {
		System::mSleep(1);
		file.open(QIODevice::ReadOnly);
		wait--;
	}
	//qWarning("wait: %d", wait);

	if (! currDir.isEmpty())
		System::chDir(currDir);

	if (!file.isOpen()) {
		qWarning("CRITICAL: Cannot open flags.lch file.");
		goto bail;
	}

	file.read((char*)(&easyFlagSize), sizeof(int));
	file.read((char*)(&flagSize), sizeof(int));
	if (easyFlagSize != sizeof(EasyFlag) || flagSize != sizeof(Flag)) {
		qWarning("CRITICAL: fs2_open and launcher versions don't match.");
		goto bail;
	}

	file.read((char*)(&numEasyFlags), sizeof(int));
	easyFlags = new EasyFlag[numEasyFlags];
	if (easyFlags == NULL) {
		qWarning("CRITICAL: Cannot allocate memory.");
		goto bail;
	}
	file.read((char*)easyFlags, sizeof(EasyFlag) * numEasyFlags);

	file.read((char*)(&numFlags), sizeof(int));
	exeParams = new Flag[numFlags];
	if (exeParams == NULL) {
		qWarning("CRITICAL: Cannot allocate memory.");
		goto bail;
	}
	file.read((char*)exeParams, sizeof(Flag) * numFlags);

	/* lets rewrite those ugly :-) C structs to something more friendly */
	for (int i = 0; i < numFlags; i++) {
		LFlag n;
		n.flag = exeParams[i].name;
		n.desc = exeParams[i].desc;
		n.onFlags = exeParams[i].on_flags;
		n.offFlags = exeParams[i].off_flags;
		n.type = exeParams[i].type;
		n.url = exeParams[i].web_url;
		n.state = false;
		if(exeTypes.indexOf(n.type) == -1)
			exeTypes.append(n.type);
		exeFlags.append(n);
	}
	for (int i = 0; i < numEasyFlags; i++)
		easyTypes.append(easyFlags[i].name);

	//qWarning("num easy: %d num flags: %d", numEasyFlags, numFlags);
	//qWarning("easy size: %d size: %d", easyFlagSize, flagSize);

	delete easyFlags;
	delete exeParams;
	file.close();
	file.remove();
	return true;

bail:
	if(easyFlags) delete easyFlags;
	if(exeParams) delete exeParams;
	file.close();
	file.remove();
	return false;
}

bool Settings::setModName(const QString &modName)
{
	this -> modName = modName;
	mods.clear();
	modImage = modInfo = modWebsite = modForum = "";

	QString tmp;
	QStringList tmpList;
	IniParser modIni;
	if (! modIni.loadIni(FilePaths::getModIni(modName)))
		return false;

	modImage = modIni.getEntry("launcher:image255x112");
	modInfo = modIni.getEntry("launcher:infotext");
	modWebsite = modIni.getEntry("launcher:website");
	modForum = modIni.getEntry("launcher:forum");

	/* what about settings.ini, should we care? */
	tmp = modIni.getEntry("settings:flags");
	if (!tmp.isEmpty()) {
		tmpList = tmp.split(" ");
		for (int i = 0; i < tmpList.size(); i++)
			setFlag(tmpList.at(i), true);
	}

	tmp = modIni.getEntry("multimod:primarylist");
	tmpList.clear();
	tmpList = tmp.split(",");
	for (int i = 0; i < tmpList.size(); i++)
		if (! tmpList.at(i).isEmpty())
			mods.append(tmpList.at(i));

	tmp = modIni.getEntry("multimod:secondrylist");
	tmpList.clear();
	tmpList = tmp.split(",");
	for (int i = 0; i < tmpList.size(); i++)
		if (! tmpList.at(i).isEmpty())
			mods.append(tmpList.at(i));

	return true;
}

void Settings::setFlag(const QString &flag, int state)
{
	int i;
	for (i = 0; i < exeFlags.size(); i++)
		if (exeFlags.at(i).flag == flag)
			break;

	if (i < exeFlags.size()) {
		exeFlags[i].state = state;
	}
	/* 
	 * for parsing ini file, add custom flags, we wont be using that
	 * function later for this purpose but setCustomFlags() instead
	 */
	else if (state == true) {
		customFlags.remove(flag);
		customFlags.append(" ");
		customFlags.append(flag);
		customFlags = customFlags.simplified();
	}
}

void Settings::setCustomFlags(const QString &customFlags)
{
	this -> customFlags = customFlags.simplified();
}

QString Settings::genArgv(bool withExe, bool withMod)
{
	QString args;
	if (withExe)
		args = exeFilePath;

	if (withMod)
		if (! modName.isEmpty()) {
			args.append(" -mod ");
			args.append(modName);
			for(int i = 0 ; i < mods.size(); i++) {
				args.append(",");
				args.append(mods.at(i));
			}
			args.append(" ");
		}

	for (int i = 0; i < exeFlags.size(); i++)
		if (exeFlags.at(i).state != 0) {
			args.append(" ");
			args.append(exeFlags.at(i).flag);
		}

	args.append(" ");
	args.append(customFlags);

	return args.simplified();
}

void Settings::setEasyType(int type)
{
	if (type == 0)
		return;

	int flag = 0;
	switch(type) {
		case 1:
			flag = EASY_DEFAULT;
			break;
		case 2:
			flag = EASY_ALL_ON;
			break;
		case 3:
			flag = EASY_MEM_ON;
			break;
		case 4:
			flag = EASY_MEM_OFF;
			break;
	}

	for (int i = 0; i < exeFlags.size(); i++) {
		if ((exeFlags.at(i).onFlags & flag) == flag)
			exeFlags[i].state = true;
		if ((exeFlags.at(i).offFlags & flag) == flag)
			exeFlags[i].state = false;
	}

}

bool Settings::save()
{
	IniParser ini;
	ini.addEntry("launcher:exe_filepath", exeFilePath, true);
	ini.addEntry("launcher:game_flags", genArgv(false, false), true);
	ini.addEntry("launcher:active_mod", modName, true);
	if (! ini.dumpIni(FilePaths::getLauncherIni())) {
		qWarning("CRITICAL: saving ini failed.");
		return false;
	}
	QFile cmdLine(FilePaths::getCmdLine());
	if (! cmdLine.open(QIODevice::WriteOnly | QIODevice::Text)) {
		qWarning("CRITICAL: saving cmd line failed.");
		return false;
	}
	QTextStream out(&cmdLine);
	out << genArgv(false, true) << endl;

	cmdLine.close();

	return true;
}
